import inspect
import json
from dataclasses import asdict, dataclass, field

READ_ONLY_FIELDS = set(["id"])


@dataclass
class TurbineModelSpecs:
    id: int = field(default=None)
    name: str = field(default=None)
    hub_height: float = field(default=None)
    rotor_diameter: float = field(default=None)
    rated_power: float = field(default=None)
    reference_density: float = field(default=None)
    reference_turbulence_intensity: float = field(default=None)
    reference_windspeed: list[float] = field(default=None)
    thrust_coefficient: list[float] = field(default=None)
    power: list[float] = field(default=None)
    public: bool = field(default_factory=bool)

    @classmethod
    def from_dict(cls, env):
        return cls(
            **{k: v for k, v in env.items() if k in inspect.signature(cls).parameters}
        )

    def _get_api_params(self):
        params = asdict(self)
        _ = [params.pop(key, None) for key in READ_ONLY_FIELDS]
        return params

    def __str__(self) -> str:
        return f"{self.__class__}:\n{json.dumps(asdict(self), indent=4)}"

    def __repr__(self) -> str:
        return json.dumps(asdict(self))



from pathlib import Path
from typing import Union

from whiffle_client.wind.components.turbine_model_specs import TurbineModelSpecs
from whiffle_client.decorators import load_data
from whiffle_client.wind.mapping.base import BaseMapping


class TurbineModelSpecsEndpoints(BaseMapping):

    URL = "/api/aspforge/turbine-models"
    RESOURCE_TYPE = "TurbineModelSpecs"

    # Turbine models commands
    @load_data(TurbineModelSpecs)
    def add(
        self,
        data: Union[str, dict, Path, TurbineModelSpecs] = None,
    ) -> TurbineModelSpecs:
        """Add new Turbine Model

        Parameters
        ----------
        data : Union[str, dict, Path, TurbineModelSpecs], optional
            Either a path to a yaml, the data itself as a dictionary or a TurbineModelSpecs object
            containing the parameters that define the turbine model specs, by default None

        Returns
        -------
        TurbineModelSpecs
            Object instance of the turbine model
        """
        request = self.session.post_request(
            f"{self.session.server_url}{self.URL}", data=data
        )
        return TurbineModelSpecs.from_dict(request.json())

    @load_data(TurbineModelSpecs)
    def edit(
        self,
        turbine_model_id: str,
        data: Union[str, dict, Path, TurbineModelSpecs] = None,
    ) -> TurbineModelSpecs:
        """Edit existing Turbine Model

        Parameters
        ----------
        turbine_model_id : str
            Id of the turbine model
        data : Union[str, dict, Path, TurbineModelSpecs], optional
            Either a path to a yaml, the data itself as a dictionary or a TurbineModelSpecs object
            containing the parameters that define the turbine model specs, by default None

        Returns
        -------
        TurbineModelSpecs
            Object instance of the turbine model
        """
        request = self.session.put_request(
            f"{self.session.server_url}{self.URL}/{turbine_model_id}",
            data=data,
        )
        return TurbineModelSpecs.from_dict(request.json())

    def get_all(self) -> list[TurbineModelSpecs]:
        """Get a list of all the Turbine Models available to the user

        Returns
        -------
        list[TurbineModelSpecs]
            List of TurbineModelSpecs object instances
        """
        request = self.session.get_request(f"{self.session.server_url}{self.URL}")
        return [
            TurbineModelSpecs.from_dict(turbine_model_specs)
            for turbine_model_specs in request.json()
        ]

    def get(
        self,
        turbine_model_id: str = None,
        turbine_model_name: str = None,
    ) -> Union[TurbineModelSpecs, list[TurbineModelSpecs]]:
        """Get a Turbine Model

        Parameters
        ----------
        turbine_model_id : int, optional
            Id of the turbine model, by default None
        turbine_model_name : str, optional
            Turbine model name, by default None

        Returns
        -------
        Union[TurbineModelSpecs, list[TurbineModelSpecs]]
            Uniquely identidied TurbineModelSpecs object instance or list of TurbineModelSpecs

        Raises
        ------
        ValueError
            Raises error if either `turbine_model_id` or `turbine_model_name` are not provided
        """
        if turbine_model_id:
            return TurbineModelSpecs.from_dict(
                self.session.get_request(
                    f"{self.session.server_url}{self.URL}/{turbine_model_id}"
                ).json()
            )
        elif turbine_model_name:
            print(
                "WARN: Accessing turbine model specs by name. Using an `id` is recomended as name does not ensure uniqueness."
            )
            return [
                TurbineModelSpecs.from_dict(specs)
                for specs in self.session.get_request(
                    f"{self.session.server_url}{self.URL}?name={turbine_model_name}"
                ).json()
            ]
        else:
            raise ValueError(
                "Please provide either valid turbine_model_id or turbine_model_name"
            )

    def delete(
        self,
        turbine_model_id: str,
    ) -> TurbineModelSpecs:
        """Delete a Turbine Model

        Parameters
        ----------
        turbine_model_id : str
            Id of the turbine model

        Returns
        -------
        TurbineModelSpecs
            TurbineModelSpecs object instance of deleted turbine model
        """
        return TurbineModelSpecs.from_dict(
            self.session.delete_request(
                f"{self.session.server_url}{self.URL}/{turbine_model_id}"
            ).json()
        )

    def download(self):
        """Not implemented"""
        return super().download()



from whiffle_client.wind.mapping.base import BaseMapping
from whiffle_client.wind.mapping.turbine_model_specs import TurbineModelSpecsEndpoints
from whiffle_client.wind.mapping.wind_simulation_tasks import (
    WindSimulationTaskEndpoints,
)



import collections
import warnings

import numpy as np
from sklearn.metrics import r2_score


def get_metric_name(metric_name):
    names = {
        "npoints": "# of data points",
        "mean(a)": "mean(a)",
        "mean(b)": "mean(b)",
        "bias": "bias",
        "r2": r"$R^2$",
        "pearsonr": r"$\rho$",
        "nmae": "nMAE",
        "mae": "MAE",
        "rmse": "RMSE",
        "nrmse": "nRMSE",
        "rmsd": "RMSD",
        "nrmsd": "nRMSD",
    }
    return names[metric_name]


def get_metric(metric_name):
    return globals()[metric_name]


def normalize(x, axis=None, norm=None):
    if norm is None:
        return 1.0
    elif norm == "mean":
        return np.nanmean(x, axis=axis)
    elif isinstance(norm, float):
        return norm
    else:
        raise ValueError("expected 'mean' or float")


def rmsd(x, y, axis=None, norm=None):
    y_bias = bias(x, y, axis=axis)

    y_minus_bias = y - y_bias

    return rmse(x, y_minus_bias, axis=axis) / normalize(x, axis=axis, norm=norm)


def rmse(x, y, axis=None, norm=None):
    return np.sqrt(((x - y) ** 2).mean()) / normalize(x, axis=axis, norm=norm)


def bias(x, y, axis=None, norm=None):
    return (y.mean(axis=axis) - x.mean(axis=axis)) / normalize(x, axis=axis, norm=norm)


def mae(x, y, axis=None, norm=None):
    return np.nanmean(np.abs(x - y), axis=axis) / normalize(x, axis=axis, norm=norm)


def r2(x, y, **kw):
    """R^2 (coefficient of determination) regression score function.

    See
    https://scikit-learn.org/stable/modules/generated/sklearn.metrics.r2_score.html
    and
    https://scikit-learn.org/stable/modules/model_evaluation.html#r2-score
    """
    return r2_score(x, y, **kw)


def pearsonr(x, y, axis=0):
    return x.to_frame().corrwith(y.to_frame()).squeeze()


def skill(a, b, norm=None):
    if norm is None:
        return collections.OrderedDict(
            [
                ("npoints", a.size),
                ("mean(a)", np.nanmean(a)),
                ("mean(b)", np.nanmean(b)),
                ("bias", bias(a, b)),
                ("pearsonr", pearsonr(a, b)),
                ("rmse", rmse(a, b)),
                ("rmsd", rmsd(a, b)),
                ("mae", mae(a, b)),
            ]
        )
    else:
        return collections.OrderedDict(
            [
                ("npoints", a.size),
                ("mean(a)", np.nanmean(a)),
                ("mean(b)", np.nanmean(b)),
                ("bias", bias(a, b)),
                ("pearsonr", pearsonr(a, b)),
                ("rmse", rmse(a, b)),
                ("nrmse", rmse(a, b, norm=norm)),
                ("rmsd", rmsd(a, b)),
                ("nrmsd", rmsd(a, b, norm=norm)),
                ("mae", mae(a, b)),
                ("nmae", mae(a, b, norm=norm)),
            ]
        )


def skill_str(a, b, metrics=None, norm=None, fmt=None):
    """Create a skill-string that is easily parsed by the 'ax.text' method.

    Parameters
    ----------
    a: array-like
        The observation.
    b: array-like
        The model.
    metrics: list or dict-like
        List of metrics or mapping of metric to metric name.
    norm:
        Norm passed to underlying skill functions.
    fmt: string or dict-like
        String format of the metrics in `metrics`, e.g.
        '.4f' or {'mean': '.0f', pearsonr: '.3f'}

    Examples
    --------
    ax.text(0.05, 0.95,
            skill_str(a, b),
            transform=ax.transAxes,
            fontsize=6,
            verticalalignment='top',
            bbox=dict(boxstyle='round', facecolor='white'),
            )
    """
    if metrics is None:
        if norm is None:
            metrics = [
                "npoints",
                "mean(a)",
                "mean(b)",
                "bias",
                "pearsonr",
                "rmse",
                "rmsd",
                "mae",
            ]
        else:
            metrics = [
                "npoints",
                "mean(a)",
                "mean(b)",
                "bias",
                "pearsonr",
                "nrmse",
                "nrmsd",
                "nmae",
            ]

    if fmt is None:
        fmt = ".5g"

    if isinstance(fmt, str):
        fmt = {metric: fmt for metric in metrics}

    if isinstance(metrics, list):
        metrics = collections.OrderedDict(
            [(metric, get_metric_name(metric)) for metric in metrics]
        )

    skill_dict = skill(a, b, norm)

    msg = ""
    for metric, metric_name in metrics.items():
        value = skill_dict[metric]
        try:
            msg += "{} = {:{sfmt:s}}\n".format(metric_name, value, sfmt=fmt[metric])
        except ValueError as err:
            # if formatting fails format as string
            warnings.warn(
                "failed to format {} as '{}': {}".format(value, fmt[metric], str(err))
            )
            msg += "{} = {:{sfmt:s}}\n".format(metric_name, value, sfmt="s")

    return msg.strip()



CHUNK_SIZE = 1024 * 1024
VALID_DOWNLOAD_FILES = ["fields", "time_series", "results"]


def download_write_chunks(filename, res):
    file = open(filename, "wb")
    file_size = float(res.headers["Content-Length"])
    downloaded = 0
    for chunk in res.iter_content(chunk_size=CHUNK_SIZE):
        if chunk:
            file.write(chunk)
            downloaded += len(chunk)
            print(
                "Downloaded {:>5.1%}\r".format(downloaded / file_size),
                end="",
                flush=True,
            )
    file.close()



from dataclasses import dataclass, field


@dataclass
class Geometries:
    les_domain: dict = field(default=None)
    meso_domain: dict = field(default=None)
    suggested_les_domain: dict = field(default=None)



from dataclasses import dataclass, field

from whiffle_client.wind.components.turbine import Turbine


@dataclass
class Windfarm:
    name: str = field(default=None)
    turbines: list[Turbine] = field(default=None)
    include_in_les: bool = field(default=True)
    thrust: bool = field(default=True)
    turbine_model_name: str = field(default=None)



#!/usr/bin/env python3

from whiffle_client.wind.mapping import (
    TurbineModelSpecsEndpoints,
    WindSimulationTaskEndpoints,
)
from whiffle_client.base_client import BaseClient


class WindSimulationClient:
    """
    Client to connect to the Whiffle Wind API
    """

    # NOTE: https://refactoring.guru/design-patterns/facade
    def __init__(self, access_token=None, url=None, session=None):
        self.session = BaseClient(access_token, url, session)

        # Map the turbine model specs endpoints
        self.turbine_model = TurbineModelSpecsEndpoints(self.session)
        self._map_turbine_model_specs_methods()

        # Map the wind simulation model endpoints
        self.wind_simulation_task = WindSimulationTaskEndpoints(self.session)
        self._map_wind_simulation_model_methods()

    def __repr__(self) -> str:
        return self.session.__repr__()

    def _map_turbine_model_specs_methods(self):
        self.add_turbine_model = self.turbine_model.add
        self.edit_turbine_model = self.turbine_model.edit
        self.get_all_turbine_models = self.turbine_model.get_all
        self.get_turbine_model = self.turbine_model.get
        self.delete_turbine_model = self.turbine_model.delete

    def _map_wind_simulation_model_methods(self):
        self.add_wind_simulation_task = self.wind_simulation_task.add
        self.edit_wind_simulation_task = self.wind_simulation_task.edit
        self.get_all_wind_simulation_tasks = self.wind_simulation_task.get_all
        self.get_wind_simulation_task = self.wind_simulation_task.get
        self.delete_wind_simulation_task = self.wind_simulation_task.delete
        self.download_wind_simulation_task = self.wind_simulation_task.download
        self.submit_wind_simulation_task = self.wind_simulation_task.submit



import numpy as np
import pandas as pd
import xarray as xr
from matplotlib import pyplot as plt

from whiffle_client.analysis.skill import skill, skill_str
from whiffle_client.analysis.utils import (
    mean_wind_speed_by_hour_of_the_day,
    weibull_params_from_samples,
    weibull_pdf_from_params,
    wind_rose_frequency,
    wind_speed_histogram,
)


def _concat_wind_data(
    measured_wind_speed_series,
    aspire_wind_speed_series,
    concurrent,
):
    wind_speed_data = [
        measured_wind_speed_series.rename("measured"),
        aspire_wind_speed_series.rename("whiffle-wind"),
    ]
    wind_speed_df = pd.concat(wind_speed_data, axis=1)
    if concurrent:
        wind_speed_df = wind_speed_df.dropna()
    return wind_speed_df


def _plot_scatter_density_hist(
    x,
    y,
    ax=None,
    add_skill_str=True,
    fancy_skill_str=False,
    cmap=plt.cm.inferno,
    one_to_one=True,
    normalize=1.0,
):
    def _hist_colouring(
        xarray, yarray, nbins=1000, cmap=plt.cm.inferno, smoother=15, fnc=None
    ):
        """Helper function to generate a colored histogram."""

        def bin_indices(data, bins):
            ixs = np.zeros(len(data), int) - 1
            for i, bini in enumerate(bins[:-1]):
                inbin = (bini <= data) * (data < bins[i + 1])
                if i == len(bins) - 1:
                    inbin *= data == bins[i + 1]
                ixs += inbin * (i + 1)
            return ixs

        xarray = xarray[np.invert(np.isnan(xarray) | np.isnan(yarray))]
        yarray = yarray[np.invert(np.isnan(xarray) | np.isnan(yarray))]

        xbins = np.linspace(
            np.floor(np.min(xarray)) - 0.1, np.ceil(np.max(xarray)) + 0.1, nbins + 1
        )
        ybins = np.linspace(
            np.floor(np.min(yarray)) - 0.1, np.ceil(np.max(yarray)) + 0.1, nbins + 1
        )
        hist2d, _, _ = np.histogram2d(xarray, yarray, bins=[xbins, ybins])
        if fnc is not None:
            hist2d = fnc(hist2d)
        x_ix = bin_indices(xarray, xbins)
        y_ix = bin_indices(yarray, ybins)

        if smoother:
            smoothhist = np.zeros(len(x_ix))
            for i, (ix, iy) in enumerate(zip(x_ix, y_ix)):
                # smoothhist[i] = np.sum(hist2d[ix-smoother:ix+smoother+1,iy-smoother:iy+smoother+1])
                smoothhist[i] = np.mean(
                    hist2d[
                        max(0, ix - smoother) : ix + smoother + 1,
                        max(0, iy - smoother) : iy + smoother + 1,
                    ]
                )
            return cmap(smoothhist / np.max(smoothhist))
        else:
            return cmap(hist2d[x_ix, y_ix] / np.max(hist2d))

    inds = np.isfinite(x) & np.isfinite(y)

    x = x[inds]
    y = y[inds]

    cvals = _hist_colouring(x, y, nbins=1000, cmap=cmap, smoother=15, fnc=None)

    ax.scatter(x, y, c=cvals, edgecolor=None, s=1)

    my_skill_names = {"bias": "bias", "corr2": r"$R^2$", "MAE": "MAE", "RMSE": "RMSE"}

    def my_skill_str(a, b, skill_names, norm=None):
        msg = ""

        msg += "# of data points = {:d}\n".format(a.size)

        msg += "mean(a) = {:.4f}\n".format(float(np.nanmean(a)))
        msg += "mean(b) = {:.4f}\n".format(float(np.nanmean(b)))

        for score, value in skill(a, b, norm).items():
            if score in skill_names:
                msg += "{} = {:.4f}\n".format(skill_names[score], float(value))

        return msg.strip()

    if add_skill_str:
        if fancy_skill_str:
            skill_str_value = my_skill_str(
                x, y, my_skill_names, normalize
            )  # y is model
            props = dict(boxstyle="round", facecolor="white")

            ax.text(
                0.05,
                0.95,
                skill_str_value,
                transform=ax.transAxes,
                fontsize=10,
                verticalalignment="top",
                bbox=props,
            )

        else:
            skill_str_value = skill_str(y, x, norm=normalize)  # y is model

            props = dict(boxstyle="round", facecolor="white")

            ax.text(
                0.05,
                0.95,
                skill_str_value,
                transform=ax.transAxes,
                fontsize=6,
                verticalalignment="top",
                bbox=props,
            )

    if one_to_one:
        start = min(np.nanmin(x), np.nanmin(y))
        end = max(np.nanmax(x), np.nanmax(y))

        ax.plot([start, end], [start, end], "-k", zorder=4)

        ax.set_xlim(start, end)
        ax.set_ylim(start, end)

        ax.set_aspect(1)

    ax.set_xlabel("Measured wind speed [m/s]")
    ax.set_ylabel("Whiffle-wind wind speed [m/s]")
    ax.set_title("Wind speed scatter density plot with metrics")
    ax.grid()

    return ax


def plot_weibull_pdf_comparison(
    measured_wind_speed_series: pd.Series,
    aspire_wind_speed_series: pd.Series,
    concurrent: bool = True,
    ax: plt.axes = None,
) -> plt.axes:
    """Fit a Weibull distribution to the data and plot the PDF.

    Parameters
    ----------
    measured_wind_speed_series : pd.Series
        Measured wind speed series.
    aspire_wind_speed_series : pd.Series
        Simulated wind speed series.
    concurrent : bool, optional
        If True, compare concurrent wind speeds i.e. an intersection of
        measured and simulated data on the time index will be performed
        before plotting.
    ax : plt.axes, optional
        Matplotlib axes.

    Returns
    -------
    plt.axes
        Matplotlib axes with the plot.
    """
    if ax is None:
        _, ax = plt.subplots()
    wind_speed_df = _concat_wind_data(
        measured_wind_speed_series,
        aspire_wind_speed_series,
        concurrent,
    )
    weibull_params_df = wind_speed_df.apply(
        lambda x: pd.Series(
            weibull_params_from_samples(x),
            index=["k", "A"],
        )
    )
    x_pdf = np.linspace(0, np.ceil(wind_speed_df.max().max()), 200)
    weibull_pdf_df = weibull_params_df.apply(
        lambda x: pd.Series(
            weibull_pdf_from_params(x_pdf, *x[["k", "A"]].tolist()),
            index=x_pdf,
        )
    )
    weibull_pdf_df.columns = [
        f'{c}:\nk={weibull_params_df[c]["k"]:.2f}\nA={weibull_params_df[c]["A"]:.2f} m/s'
        for c in weibull_pdf_df.columns
    ]
    weibull_pdf_df.plot(ax=ax)
    ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    ax.set_xlabel("Wind speed [m/s]")
    ax.set_ylabel("Weibull pdf [-]")
    ax.set_title("Fitted Weibull PDF comparison on wind speed")
    ax.grid()
    return ax


def plot_wind_speed_distribution_comparison(
    measured_wind_speed_series: pd.Series,
    aspire_wind_speed_series: pd.Series,
    concurrent: bool = True,
    bin_width: float = 1.0,
    kind: str = "bar",
    ax: plt.Axes = None,
) -> plt.Axes:
    """Compare wind speed distributions between measured and simulated data.

    Notes
    -----
    When 'kind' parameter is set to 'line', each point correspond to the
    histogram bin frequency (not the density). The 'line' plot represents
    the same underlying data as the 'bar' plot but is helpful for comparison.

    Parameters
    ----------
    measured_wind_speed_series : pd.Series
        Measured wind speed series.
    aspire_wind_speed_series : pd.Series
        Simulated wind speed series.
    concurrent : bool, optional
        If True, compare concurrent wind speeds i.e. an intersection of
        measured and simulated data on the time index will be performed
        before plotting.
    bin_width : float, optional
        Binning width.
    kind : str, optional
        Whether to plot distributions as 'bar' or 'line'.
    ax : plt.Axes, optional
        Matplotlib axes.

    Returns
    -------
    plt.Axes
        Matplotlib axes with the plot.
    """
    allowed_kind = ["bar", "line"]
    if kind not in allowed_kind:
        raise ValueError(f"'kind' parameter should be in {allowed_kind}")
    if ax is None:
        _, ax = plt.subplots()
    wind_speed_df = _concat_wind_data(
        measured_wind_speed_series,
        aspire_wind_speed_series,
        concurrent,
    )
    ws_hist_df = wind_speed_df.apply(wind_speed_histogram, bin_width=bin_width)
    ws_hist_df = ws_hist_df.fillna(0)
    ws_hist_df.index = ws_hist_df.index.map(lambda x: x.mid)
    ws_hist_df = ws_hist_df / ws_hist_df.sum()
    if kind == "line":
        ws_hist_df.plot(marker=".", ax=ax, legend=True)
    elif kind == "bar":
        ws_hist_df.plot.bar(ax=ax, legend=True)
    ax.set_xlabel("Wind speed bin center [m/s]")
    ax.set_ylabel("Frequency [-]")
    ax.set_title("Wind speed histogram comparison")
    ax.grid()
    return ax


def plot_wind_rose_frequency_comparison(
    measured_wind_direction_series: pd.Series,
    aspire_wind_direction_series: pd.Series,
    concurrent: bool = True,
    n_sectors: int = 12,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Compare sector frequencies between measured and simulated data.

    Notes
    -----
    This plot is in polar projection.

    Parameters
    ----------
    measured_wind_direction_series : pd.Series
        Measured wind direction series.
    aspire_wind_direction_series : pd.Series
        Simulated wind direction series.
    concurrent : bool, optional
        If True, compare concurrent wind directions i.e. an intersection of
        measured and simulated data on the time index will be performed
        before plotting.
    n_sectors : int, optional
        Number of sectors to use for the wind rose.
    ax : plt.Axes, optional
        Matplotlib axes.

    Returns
    -------
    plt.Axes
        Matplotlib axes with the plot.
    """
    if ax is None:
        _, ax = plt.subplots(subplot_kw={"projection": "polar"})
    wind_dir_df = _concat_wind_data(
        measured_wind_direction_series,
        aspire_wind_direction_series,
        concurrent,
    )
    wind_rose_df = wind_dir_df.apply(
        wind_rose_frequency,
        n_sectors=n_sectors,
    )
    wind_rose_df = wind_rose_df / wind_rose_df.sum()
    wind_rose_df_rad = wind_rose_df.copy()
    wind_rose_df_rad.index = wind_rose_df_rad.index.map(lambda x: np.deg2rad(x.mid))
    for c in wind_rose_df_rad.columns:
        x = wind_rose_df_rad.index.to_series()
        x = pd.concat([x, x.iloc[[0]]]).to_numpy()
        y = wind_rose_df_rad[c]
        y = pd.concat([y, y.iloc[[0]]]).to_numpy()
        ax.plot(
            x,
            y,
            marker=".",
            label=c,
        )
    ax.set_theta_zero_location("N")
    ax.set_theta_direction(-1)
    ax.grid(True)
    ax.legend(loc="center left", bbox_to_anchor=(1.1, 0.5))
    ax.set_xlabel("Wind direction bin center [m/s]")
    ax.set_ylabel("Frequency [-]", labelpad=30)
    ax.set_title("Wind rose frequency comparison")
    return ax


def plot_wind_speed_scatter(
    measured_wind_speed_series: pd.Series,
    aspire_wind_speed_series: pd.Series,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Scatter plot of measured and simulated wind speeds.

    Parameters
    ----------
    measured_wind_speed_series : pd.Series
        Measured wind speed series.
    aspire_wind_speed_series : pd.Series
        Simulated wind speed series.
    ax : plt.Axes, optional
        Matplotlib axes.

    Returns
    -------
    plt.Axes
        Matplotlib axes with the plot.
    """
    if ax is None:
        _, ax = plt.subplots()
    wind_speed_df = _concat_wind_data(
        measured_wind_speed_series,
        aspire_wind_speed_series,
        concurrent=True,
    )
    ws_min = np.floor(wind_speed_df.min().min())
    ws_max = np.ceil(wind_speed_df.max().max())

    ax.set_aspect("equal")
    wind_speed_df.plot.scatter(
        x="measured",
        y="whiffle-wind",
        s=10,
        alpha=0.2,
        ax=ax,
    )
    ax.plot([ws_min, ws_max], [ws_min, ws_max], "-", color="k", lw=0.5)
    ax.set_xlim(ws_min, ws_max)
    ax.set_ylim(ws_min, ws_max)
    ax.set_xlabel("Measured wind speed [m/s]")
    ax.set_ylabel("Whiffle-wind wind speed [m/s]")
    ax.set_title("Wind speed scatter plot")
    ax.grid()
    return ax


def plot_wind_speed_scatter_density(
    measured_wind_speed_series: pd.Series,
    aspire_wind_speed_series: pd.Series,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Density plot of measured and simulated wind speeds.

    Notes
    -----
    This plot is particularly useful when the data contains many samples.

    Parameters
    ----------
    measured_wind_speed_series : pd.Series
        Measured wind speed series.
    aspire_wind_speed_series : pd.Series
        Simulated wind speed series.
    ax : plt.Axes, optional
        Matplotlib axes.

    Returns
    -------
    plt.Axes
        Matplotlib axes with the plot.
    """
    if ax is None:
        _, ax = plt.subplots()
    return _plot_scatter_density_hist(
        measured_wind_speed_series,
        aspire_wind_speed_series,
        ax=ax,
    )


def plot_wind_speed_diurnal_cycle_comparison(
    measured_wind_speed_series: pd.Series,
    aspire_wind_speed_series: pd.Series,
    concurrent: bool = True,
    ax: plt.Axes = None,
) -> plt.Axes:
    """Compare mean diurnal cycle of measured and simulated wind speeds.

    Notes
    -----
    Each point in the plot represents the mean wind speed grouped
    by hour of the day.

    Parameters
    ----------
    measured_wind_speed_series : pd.Series
        Measured wind speed series.
    aspire_wind_speed_series : pd.Series
        Simulated wind speed series.
    concurrent : bool, optional
        If True, compare concurrent wind speeds i.e. an intersection of
        measured and simulated data on the time index will be performed
        before plotting.
    ax : plt.Axes, optional
        Matplotlib axes.

    Returns
    -------
    plt.Axes
        Matplotlib axes with the plot.
    """
    if ax is None:
        _, ax = plt.subplots()
    wind_speed_df = _concat_wind_data(
        measured_wind_speed_series,
        aspire_wind_speed_series,
        concurrent,
    )
    wind_speed_diurnal_cycle = wind_speed_df.apply(
        mean_wind_speed_by_hour_of_the_day,
    )
    wind_speed_diurnal_cycle.plot(marker=".", ax=ax)
    ax.set_xlabel("Hour of the day [-]")
    ax.set_ylabel("Mean wind speed by hour of the day [m/s]")
    ax.set_title("Wind speed diurnal cycle comparison")
    ax.grid()
    return ax


def plot_wind_speed_vertical_profile_comparison(
    measured_wind_speed_da: xr.DataArray,
    aspire_wind_speed_da: xr.DataArray,
    concurrent: bool = True,
    ax: bool = None,
) -> plt.Axes:
    """Compare vertical profiles of measured and simulated wind speeds.

    Notes
    -----
        - Data arrays should contain time and height dimensions.
        - Data arrays can have different heights.
        - If 'concurrent' is True, time intersection between all heights
          and data arrays will be performed.

    Parameters
    ----------
    measured_wind_speed_da : xr.DataArray
        Measured wind speed data array.
    aspire_wind_speed_da : xr.DataArray
        Simulated wind speed data array.
    concurrent : bool, optional
        If True, compare concurrent wind speeds i.e. an intersection of
        measured and simulated data on the time index will be performed
        before plotting.
    ax : bool, optional
        _description_, by default None

    Returns
    -------
    plt.Axes
        _description_
    """
    if ax is None:
        _, ax = plt.subplots()
    measured_wind_speed_da = measured_wind_speed_da.dropna("height", how="all")
    if concurrent:
        measured_time_idx = measured_wind_speed_da.dropna("time")["time"].to_index()
        aspire_time_idx = aspire_wind_speed_da.dropna("time")["time"].to_index()
        concurrent_time_idx = measured_time_idx.intersection(aspire_time_idx)
        measured_wind_speed_da = measured_wind_speed_da.sel(time=concurrent_time_idx)
        aspire_wind_speed_da = aspire_wind_speed_da.sel(time=concurrent_time_idx)
    measured_wind_speed_da.mean("time", keep_attrs=True).plot(
        y="height",
        label="measured",
        marker=".",
        ax=ax,
    )
    aspire_wind_speed_da.mean("time", keep_attrs=True).plot(
        y="height",
        label="whiffle-wind",
        marker=".",
        ax=ax,
    )
    ax.legend()
    ax.set_xlabel("Mean wind speed [m/s]")
    ax.set_title("Mean wind speed vertical profile comparison")
    ax.grid()
    return ax



import yaml

from tests.conftest import (
    mock_new_task_response,
    get_default_task,
    default_task_path,
    task_with_yaml_include_path,
)
from whiffle_client import Client


class TestAPI:
    def test_init_client_with_default_params(self):
        config = Client.get_config()
        client = Client()
        assert client.server_url == config["whiffle"]["url"]
        assert config["user"]["token"] in client.session.headers["Authorization"]

    def test_init_client_with_given_params(self):
        token = "test_token"
        url = "https://test_url.com"
        client = Client(token, url)

        assert client.server_url == url
        assert token in client.session.headers["Authorization"]

    def test_new_task_from_dict(self, mocker):
        task = get_default_task()
        client = Client()
        mock_new_task_response(mocker)
        task_id = client.new_task(task)
        assert type(task_id) == str

    def test_new_task_from_json(self, mocker):
        client = Client()
        mock_new_task_response(mocker)
        task_id = client.new_task(default_task_path)
        assert type(task_id) == str

    def test_new_task_from_yaml(self, mocker, tmp_path):
        task = get_default_task()
        path = f"{tmp_path}/task.yaml"
        with open(path, "w") as f:
            yaml.safe_dump(task, f)
        client = Client()
        mock_new_task_response(mocker)
        task_id = client.new_task(path)
        assert type(task_id) == str

    def test_new_task_from_yaml_with_include(self, mocker, tmp_path):
        task = get_default_task()
        client = Client()
        mock_new_task_response(mocker)
        task_id = client.new_task(task_with_yaml_include_path)
        assert type(task_id) == str

    def test_new_task_with_warnings(self, mocker, capsys):
        warnings = {"version": "incorrect version"}

        task = get_default_task()
        client = Client()
        mock_new_task_response(mocker, warnings)
        client.new_task(task)
        captured = capsys.readouterr()
        assert "incorrect version" in captured.out



import datetime
import inspect
import json
from dataclasses import asdict, dataclass, field

import yaml
from yamlinclude import YamlIncludeConstructor

from whiffle_client.loaders import csv_loader_constructor
from whiffle_client.wind.components import Domain, Geometries, Metmast, Windfarm
from whiffle_client.wind.components.dates import Dates

YamlIncludeConstructor.add_to_loader_class(loader_class=yaml.FullLoader)
yaml.add_constructor("!include-csv", csv_loader_constructor)

READ_ONLY_FIELDS = set(
    [
        "id",
        "status",
        "messages",
        "creator",
        "created",
        "updated",
        "finished",
        "test_run",
        "price",
        "price_overview",
        "reference_number",
        "configuration_valid",
        "geometries",
        "domain",
        "progress",
    ]
)


@dataclass
class WindSimulationTask:
    # Metadata attributes
    id: str = field(default=None)
    status: str = field(default=None)
    messages: list[dict] = field(default=None)
    creator: dict = field(default=None)

    created: datetime.datetime = field(default=None)
    updated: datetime.datetime = field(default=None)
    finished: datetime.datetime = field(default=None)

    price: float = field(default=None)
    price_overview: dict = field(default=None)

    reference_number: str = field(default=None)
    reference_code: str = field(default=None)
    name: str = field(default=None)
    configuration_valid: bool = field(default=None)
    test_run: bool = field(default=None)

    # Time attributes
    dates: Dates = field(default=None)

    # Domain/object/geolocated elements attributes
    windfarms: list[Windfarm] = field(default_factory=list)
    geometries: Geometries = field(default=None)
    domain: Domain = field(default=None)
    metmasts: list[Metmast] = field(default_factory=list)
    metmasts_heights: list[float] = field(default_factory=list)

    # Output attributes
    wind_resource_grid_heights: list[float] = field(default_factory=list)
    fields_heights: list[float] = field(default_factory=list)

    progress: dict = field(default=None)

    @classmethod
    def from_dict(cls, env):
        return cls(
            **{k: v for k, v in env.items() if k in inspect.signature(cls).parameters}
        )

    def __post_init__(self):

        if isinstance(self.metmasts, list):
            self.metmasts = [
                Metmast(
                    id=metmast["name"],
                    longitude=metmast["longitude"],
                    latitude=metmast["latitude"],
                )
                for metmast in self.metmasts
            ]
        elif isinstance(self.metmasts, dict):
            self.metmasts = [
                Metmast(id=name, latitude=latitude, longitude=longitude)
                for name, (longitude, latitude) in zip(
                    self.metmasts["name"], self.metmasts["location"]
                )
            ]

        if isinstance(self.windfarms, list):
            self.windfarms = [Windfarm(**windfarm) for windfarm in self.windfarms]
        elif isinstance(self.windfarms, dict):
            self.windfarms = [
                Windfarm(
                    name=windfarm_name,
                    include_in_les=windfarm["include_in_les"],
                    thrust=windfarm["thrust"],
                    turbine_model_name=windfarm["turbine_model_name"],
                    turbines=[
                        {
                            "name": name,
                            "longitude": longitude,
                            "latitude": latitude,
                        }
                        for name, (longitude, latitude) in zip(
                            windfarm["name"], windfarm["location"]
                        )
                    ],
                )
                for windfarm_name, windfarm in self.windfarms.items()
            ]

    def _get_api_params(self):
        params = asdict(self)
        _ = [params.pop(key, None) for key in READ_ONLY_FIELDS]

        if self.metmasts:
            params["metmasts"] = [
                {
                    "name": metmast.id,
                    "longitude": metmast.longitude,
                    "latitude": metmast.latitude,
                }
                for metmast in self.metmasts
            ]

        if self.windfarms:
            params["windfarms"] = [
                {
                    "name": windfarm.name,
                    "include_in_les": windfarm.include_in_les,
                    "thrust": windfarm.thrust,
                    "turbine_model_name": windfarm.turbine_model_name,
                    "turbines": [
                        {
                            "name": turbine["name"],
                            "longitude": turbine["longitude"],
                            "latitude": turbine["latitude"],
                        }
                        for turbine in windfarm.turbines
                    ],
                }
                for windfarm in self.windfarms
            ]

        return params

    def __str__(self) -> str:
        return f"{self.__class__}:\n{json.dumps(asdict(self), indent=4)}"

    def __repr__(self) -> str:
        return json.dumps(asdict(self))



from pathlib import Path
import yaml

from functools import wraps

from yamlinclude import YamlIncludeConstructor
from whiffle_client.loaders import csv_loader_constructor

YamlIncludeConstructor.add_to_loader_class(loader_class=yaml.FullLoader)
yaml.add_constructor("!include-csv", csv_loader_constructor)


def load_data(class_type, resource_type=None):
    def wrap(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            data = kwargs.get("data", None)
            if isinstance(data, str) or isinstance(data, Path):
                # Data provided as yaml path
                data = yaml.load(open(data), Loader=yaml.FullLoader)
                data = class_type.from_dict(data)
            elif isinstance(data, dict):
                # Data directly provided as dict
                data = class_type.from_dict(data)
            if data:
                kwargs["data"] = data._get_api_params()
            else:
                raise ValueError("Please provide valid data or path to valid data")
            return func(self, *args, **kwargs)

        return wrapper

    return wrap


def request_ok(func):
    def wrapper(self, *args, **kwargs):
        request = func(self, *args, **kwargs)
        try:
            request.raise_for_status()
            return request
        except Exception:
            raise ValueError(request.json())

    return wrapper



from dataclasses import dataclass, field


@dataclass
class Domain:
    longitude: float = field(default=None)
    latitude: float = field(default=None)
    cells: list[float] = field(default=None)
    resolution: list[float] = field(default=None)
